import inspect
import warnings
from functools import partial, wraps
from typing import Any, Callable, Dict, List, Optional, Union

import asynctest

from .transports import BaseMockTransport

__all__ = ["MockTransport", "HTTPXMock"]


class MockTransport(BaseMockTransport):
    _patches: List[asynctest.mock._patch] = []
    transports: List["MockTransport"] = []
    targets = [
        "httpcore._sync.connection.SyncHTTPConnection.request",
        "httpcore._sync.connection_pool.SyncConnectionPool.request",
        "httpcore._sync.http_proxy.SyncHTTPProxy.request",
        "httpcore._async.connection.AsyncHTTPConnection.request",
        "httpcore._async.connection_pool.AsyncConnectionPool.request",
        "httpcore._async.http_proxy.AsyncHTTPProxy.request",
        "httpx._transports.asgi.ASGITransport.request",
        "httpx._transports.wsgi.WSGITransport.request",
        "httpx._transports.urllib3.URLLib3Transport.request",
    ]

    def __init__(
        self,
        assert_all_called: bool = True,
        assert_all_mocked: bool = True,
        base_url: Optional[str] = None,
        local: bool = False,
    ) -> None:
        self._local = local
        super().__init__(
            assert_all_called=assert_all_called,
            assert_all_mocked=assert_all_mocked,
            base_url=base_url,
        )

    def __call__(
        self,
        func: Optional[Callable] = None,
        assert_all_called: Optional[bool] = None,
        assert_all_mocked: Optional[bool] = None,
        base_url: Optional[str] = None,
    ) -> Union["MockTransport", Callable]:
        """
        Decorator or Context Manager.

        Use decorator/manager with parentheses for local state, or without parentheses
        for global state, i.e. shared patterns added outside of scope.
        """
        if func is None:
            # Parantheses used, branch out to new nested instance.
            # - Only stage when using local ctx `with respx.mock(...) as respx_mock:`
            # - First stage when using local decorator `@respx.mock(...)`
            #   FYI, global ctx `with respx.mock:` hits __enter__ directly
            settings: Dict[str, Any] = {
                "base_url": base_url,
                "local": True,
            }
            if assert_all_called is not None:
                settings["assert_all_called"] = assert_all_called
            if assert_all_mocked is not None:
                settings["assert_all_mocked"] = assert_all_mocked
            return self.__class__(**settings)

        # Async Decorator
        @wraps(func)
        async def async_decorator(*args, **kwargs):
            assert func is not None
            if self._local:
                kwargs["respx_mock"] = self
            async with self:
                return await func(*args, **kwargs)

        # Sync Decorator
        @wraps(func)
        def sync_decorator(*args, **kwargs):
            assert func is not None
            if self._local:
                kwargs["respx_mock"] = self
            with self:
                return func(*args, **kwargs)

        # Dispatch async/sync decorator, depening on decorated function.
        # - Only stage when using global decorator `@respx.mock`
        # - Second stage when using local decorator `@respx.mock(...)`
        return async_decorator if inspect.iscoroutinefunction(func) else sync_decorator

    def __enter__(self) -> "MockTransport":
        self.start()
        return self

    def __exit__(self, *args: Any) -> None:
        self.stop()

    async def __aenter__(self) -> "MockTransport":
        return self.__enter__()

    async def __aexit__(self, *args: Any) -> None:
        self.__exit__(*args)

    def start(self) -> None:
        """
        Register transport and start patching.
        """
        self.transports.append(self)
        self._patch()

    def stop(self, clear: bool = True, reset: bool = True) -> None:
        """
        Unregister transport and stop patching, when no registered transports left.
        """
        try:
            if self._assert_all_called:
                self.assert_all_called()
        finally:
            if clear:
                self.clear()
            if reset:
                self.reset()

            # Unregister current transport
            assert self in self.transports, "RESPX transport already stopped!"
            self.transports.remove(self)
            self._unpatch()

    @classmethod
    def _patch(cls) -> None:
        # Ensure we only patch once!
        if cls._patches:
            return

        # Start patching target transports
        for target in cls.targets:
            patch = asynctest.mock.patch(
                target, spec=True, create=True, new_callable=cls._mock,
            )
            patch.start()
            cls._patches.append(patch)

    @classmethod
    def _unpatch(cls) -> None:
        # Ensure we don't stop patching when registered transports exists
        if cls.transports:
            return

        # Stop patching HTTPX
        while cls._patches:
            patch = cls._patches.pop()
            patch.stop()

    @classmethod
    def _mock(cls, spec):
        if inspect.iscoroutinefunction(spec):

            async def request(self, *args, **kwargs):
                kwargs["pass_through"] = partial(spec, self)
                response = None
                error = None
                for transport in cls.transports:
                    try:
                        response = await transport._async_request(*args, **kwargs)
                    except AssertionError as e:
                        error = e.args[0]
                        continue
                    else:
                        break
                else:
                    assert response, error
                return response

        else:

            def request(self, *args, **kwargs):
                kwargs["pass_through"] = partial(spec, self)
                response = None
                error = None
                for transport in cls.transports:
                    try:
                        response = transport._sync_request(*args, **kwargs)
                    except AssertionError as e:
                        error = e.args[0]
                        continue
                    else:
                        break
                else:
                    assert response, error
                return response

        return request


class HTTPXMock(MockTransport):
    def __init__(self, *args: Any, **kwargs: Any) -> None:
        warnings.warn(
            "HTTPXMock() is due to be deprecated. Use MockTransport() instead.",
            DeprecationWarning,
            stacklevel=2,
        )
        super().__init__(*args, **kwargs)



import inspect
import json as jsonlib
import re
from functools import partial
from typing import (
    Any,
    AsyncIterator,
    Callable,
    Dict,
    Iterator,
    List,
    Optional,
    Pattern,
    Sequence,
    Tuple,
    Union,
)
from urllib.parse import urljoin, urlparse

import asynctest
import httpx  # TODO: Drop usage
from httpcore import AsyncByteStream, SyncByteStream
from httpx import Headers as HTTPXHeaders  # TODO: Drop usage

URL = Tuple[bytes, bytes, Optional[int], bytes]
Headers = List[Tuple[bytes, bytes]]
TimeoutDict = Dict[str, Optional[float]]
Request = Tuple[
    bytes, URL, Headers, Union[SyncByteStream, AsyncByteStream],
]
Response = Tuple[
    bytes,  # http version
    int,  # status code
    bytes,  # reason
    Headers,
    Union[SyncByteStream, AsyncByteStream],  # body
]

HeaderTypes = Union[
    HTTPXHeaders,
    Dict[str, str],
    Dict[bytes, bytes],
    Sequence[Tuple[str, str]],
    Sequence[Tuple[bytes, bytes]],
]

Regex = type(re.compile(""))
Kwargs = Dict[str, Any]
URLPatternTypes = Union[str, Pattern[str], URL]
ContentDataTypes = Union[
    bytes, str, List, Dict, Callable, Exception,
]

istype = lambda t, o: isinstance(o, t)
isregex = partial(istype, Regex)


def build_request(request: Request) -> Union[httpx.Request, Request]:
    """
    Try to re-build a httpx request from httpcore request args
    """
    try:
        from httpx import URL as _URL, Request as _Request
    except ImportError:  # pragma: no cover
        return request
    else:
        method, url, headers, stream = request
        scheme, host, port, full_path = url
        port_str = "" if port == {b"https": 443, b"http": 80}[scheme] else f":{port}"
        return _Request(
            method.decode(),
            _URL(f"{scheme.decode()}://{host.decode()}{port_str}{full_path.decode()}"),
            headers=headers or None,
            stream=stream,  # type: ignore
        )


def build_response(
    response: Optional[Response], request: Union[httpx.Request, Request]
) -> Optional[Union[httpx.Response, Response]]:
    """
    Try to re-build a httpx response from httpcore response
    """
    if response is None:
        return None
    if not isinstance(request, httpx.Request):  # pragma: no cover
        return response
    try:
        from httpx import Response as _Response
    except ImportError:  # pragma: no cover
        return response
    else:
        http_version, status_code, _, headers, stream = response
        httpx_response = _Response(
            status_code,
            http_version=http_version.decode("ascii"),
            headers=headers,
            stream=stream,  # type: ignore
            request=request,
        )
        httpx_response.read()
        return httpx_response


class ContentStream(AsyncByteStream, SyncByteStream):
    def __init__(self, content: bytes) -> None:
        self._content = content
        self.close_func = None
        self.aclose_func = None

    def __iter__(self) -> Iterator[bytes]:
        yield self._content

    async def __aiter__(self) -> AsyncIterator[bytes]:
        yield self._content


class ResponseTemplate:
    def __init__(
        self,
        status_code: Optional[int] = None,
        headers: Optional[HeaderTypes] = None,
        content: Optional[ContentDataTypes] = None,
        content_type: Optional[str] = None,
        context: Optional[Kwargs] = None,
    ) -> None:
        self.http_version = 1.1
        self.status_code = status_code or 200
        self.context = context if context is not None else {}
        self._content = content if content is not None else b""
        self._headers = HTTPXHeaders(headers) if headers else HTTPXHeaders()
        if content_type:
            self._headers["Content-Type"] = content_type

    def clone(self, context: Optional[Kwargs] = None) -> "ResponseTemplate":
        return ResponseTemplate(
            self.status_code, self._headers, self._content, context=context
        )

    @property
    def headers(self) -> HTTPXHeaders:
        if "Content-Type" not in self._headers:
            self._headers["Content-Type"] = "text/plain"
        return self._headers

    @property
    def content(self) -> bytes:
        return self.encode_content(self._content)

    @content.setter
    def content(self, content: ContentDataTypes) -> None:
        self._content = content

    @property
    async def acontent(self) -> bytes:
        content: ContentDataTypes = self._content

        # Pre-handle async content callbacks
        if callable(content) and inspect.iscoroutinefunction(content):
            content = await content(**self.context)

        return self.encode_content(content)

    def encode_content(self, content: ContentDataTypes) -> bytes:
        if callable(content):
            content = content(**self.context)

        if isinstance(content, Exception):
            raise content

        if isinstance(content, bytes):
            return content

        if isinstance(content, (list, dict)):
            content = jsonlib.dumps(content)
            if "Content-Type" not in self._headers:
                self._headers["Content-Type"] = "application/json"

        assert isinstance(content, str), "Invalid type of content"
        content = content.encode("utf-8")  # TODO: Respect charset

        return content

    @property
    def raw(self):
        stream = ContentStream(self.content)
        return (
            f"HTTP/{self.http_version}".encode("ascii"),
            self.status_code,
            b"<reason_phrase>",
            self.headers.raw,
            stream,
        )

    @property
    async def araw(self):
        stream = ContentStream(await self.acontent)
        return (
            f"HTTP/{self.http_version}".encode("ascii"),
            self.status_code,
            b"<reason_phrase>",
            self.headers.raw,
            stream,
        )


class RequestPattern:
    def __init__(
        self,
        method: Union[str, Callable],
        url: Optional[URLPatternTypes],
        response: Optional[ResponseTemplate] = None,
        pass_through: bool = False,
        alias: Optional[str] = None,
        base_url: Optional[str] = None,
    ) -> None:
        self._match_func: Optional[Callable] = None

        if callable(method):
            self.method = None
            self.url = None
            self.pass_through = None
            self._match_func = method
        else:
            self.method = method
            self.set_url(url, base=base_url)
            self.pass_through = pass_through

        self.response = response or ResponseTemplate()
        self.alias = alias
        self.stats = asynctest.mock.MagicMock()

    @property
    def called(self):
        return self.stats.called

    @property
    def call_count(self):
        return self.stats.call_count

    @property
    def calls(self):
        return [
            (request, response) for (request, response), _ in self.stats.call_args_list
        ]

    def get_url(self) -> Optional[URLPatternTypes]:
        return self._url

    def set_url(
        self, url: Optional[URLPatternTypes], base: Optional[str] = None,
    ) -> None:
        url = url or None
        if url is None:
            url = base
        elif isinstance(url, str):
            url = url if base is None else urljoin(base, url)
            parsed_url = urlparse(url)
            if not parsed_url.path:
                url = parsed_url._replace(path="/").geturl()
        elif isinstance(url, tuple):
            url = self.build_url(url)
        elif isregex(url):
            url = url if base is None else re.compile(urljoin(base, url.pattern))
        else:
            raise ValueError(
                "Request url pattern must be str or compiled regex, got {}.".format(
                    type(url).__name__
                )
            )
        self._url = url

    url = property(get_url, set_url)

    def build_url(self, parts: URL) -> str:
        scheme, host, port, full_path = parts
        port_str = "" if port == {b"https": 443, b"http": 80}[scheme] else f":{port}"
        return f"{scheme.decode()}://{host.decode()}{port_str}{full_path.decode()}"

    def match(self, request: Request) -> Optional[Union[Request, ResponseTemplate]]:
        """
        Matches request with configured pattern;
        custom matcher function or http method + url pattern.

        Returns None for a non-matching pattern, mocked response for a match,
        or input request for pass-through.
        """
        matches = False
        url_params: Kwargs = {}
        _request = build_request(request)

        if self.pass_through:
            return request

        if self._match_func:
            response = self.response.clone(context={"request": _request})
            result = self._match_func(_request, response)
            if result == _request:  # Detect pass through
                result = request
            return result

        request_method, _request_url, *_ = request
        if self.method != request_method.decode():
            return None

        request_url = self.build_url(_request_url)
        if not self._url:
            matches = True
        elif isinstance(self._url, str):
            matches = self._url == request_url
        else:
            match = self._url.match(request_url)
            if match:
                matches = True
                url_params = match.groupdict()

        if matches:
            return self.response.clone(context={"request": _request, **url_params})

        return None



from typing import Callable, Optional, Pattern, Union

from .mocks import MockTransport
from .models import ContentDataTypes, HeaderTypes, RequestPattern

mock = MockTransport(assert_all_called=False)

aliases = mock.aliases
stats = mock.stats
calls = mock.calls


def start() -> None:
    global mock
    mock.start()


def stop(clear: bool = True, reset: bool = True) -> None:
    global mock
    mock.stop(clear=clear, reset=reset)


def clear() -> None:
    global mock
    mock.clear()


def reset() -> None:
    global mock
    mock.reset()


def add(
    method: Union[str, Callable],
    url: Optional[Union[str, Pattern]] = None,
    status_code: Optional[int] = None,
    content: Optional[ContentDataTypes] = None,
    content_type: Optional[str] = None,
    headers: Optional[HeaderTypes] = None,
    pass_through: bool = False,
    alias: Optional[str] = None,
) -> RequestPattern:
    global mock
    return mock.add(
        method,
        url=url,
        status_code=status_code,
        content=content,
        content_type=content_type,
        headers=headers,
        pass_through=pass_through,
        alias=alias,
    )


def request(
    method: Union[str, Callable],
    url: Optional[Union[str, Pattern]] = None,
    status_code: Optional[int] = None,
    content: Optional[ContentDataTypes] = None,
    content_type: Optional[str] = None,
    headers: Optional[HeaderTypes] = None,
    pass_through: bool = False,
    alias: Optional[str] = None,
) -> RequestPattern:
    global mock
    return mock.request(
        method,
        url=url,
        status_code=status_code,
        content=content,
        content_type=content_type,
        headers=headers,
        pass_through=pass_through,
        alias=alias,
    )


def get(
    url: Optional[Union[str, Pattern]] = None,
    status_code: Optional[int] = None,
    content: Optional[ContentDataTypes] = None,
    content_type: Optional[str] = None,
    headers: Optional[HeaderTypes] = None,
    pass_through: bool = False,
    alias: Optional[str] = None,
) -> RequestPattern:
    global mock
    return mock.get(
        url=url,
        status_code=status_code,
        content=content,
        content_type=content_type,
        headers=headers,
        pass_through=pass_through,
        alias=alias,
    )


def post(
    url: Optional[Union[str, Pattern]] = None,
    status_code: Optional[int] = None,
    content: Optional[ContentDataTypes] = None,
    content_type: Optional[str] = None,
    headers: Optional[HeaderTypes] = None,
    pass_through: bool = False,
    alias: Optional[str] = None,
) -> RequestPattern:
    global mock
    return mock.post(
        url=url,
        status_code=status_code,
        content=content,
        content_type=content_type,
        headers=headers,
        pass_through=pass_through,
        alias=alias,
    )


def put(
    url: Optional[Union[str, Pattern]] = None,
    status_code: Optional[int] = None,
    content: Optional[ContentDataTypes] = None,
    content_type: Optional[str] = None,
    headers: Optional[HeaderTypes] = None,
    pass_through: bool = False,
    alias: Optional[str] = None,
) -> RequestPattern:
    global mock
    return mock.put(
        url=url,
        status_code=status_code,
        content=content,
        content_type=content_type,
        headers=headers,
        pass_through=pass_through,
        alias=alias,
    )


def patch(
    url: Optional[Union[str, Pattern]] = None,
    status_code: Optional[int] = None,
    content: Optional[ContentDataTypes] = None,
    content_type: Optional[str] = None,
    headers: Optional[HeaderTypes] = None,
    pass_through: bool = False,
    alias: Optional[str] = None,
) -> RequestPattern:
    global mock
    return mock.patch(
        url=url,
        status_code=status_code,
        content=content,
        content_type=content_type,
        headers=headers,
        pass_through=pass_through,
        alias=alias,
    )


def delete(
    url: Optional[Union[str, Pattern]] = None,
    status_code: Optional[int] = None,
    content: Optional[ContentDataTypes] = None,
    content_type: Optional[str] = None,
    headers: Optional[HeaderTypes] = None,
    pass_through: bool = False,
    alias: Optional[str] = None,
) -> RequestPattern:
    global mock
    return mock.delete(
        url=url,
        status_code=status_code,
        content=content,
        content_type=content_type,
        headers=headers,
        pass_through=pass_through,
        alias=alias,
    )


def head(
    url: Optional[Union[str, Pattern]] = None,
    status_code: Optional[int] = None,
    content: Optional[ContentDataTypes] = None,
    content_type: Optional[str] = None,
    headers: Optional[HeaderTypes] = None,
    pass_through: bool = False,
    alias: Optional[str] = None,
) -> RequestPattern:
    global mock
    return mock.head(
        url=url,
        status_code=status_code,
        content=content,
        content_type=content_type,
        headers=headers,
        pass_through=pass_through,
        alias=alias,
    )


def options(
    url: Optional[Union[str, Pattern]] = None,
    status_code: Optional[int] = None,
    content: Optional[ContentDataTypes] = None,
    content_type: Optional[str] = None,
    headers: Optional[HeaderTypes] = None,
    pass_through: bool = False,
    alias: Optional[str] = None,
) -> RequestPattern:
    global mock
    return mock.options(
        url=url,
        status_code=status_code,
        content=content,
        content_type=content_type,
        headers=headers,
        pass_through=pass_through,
        alias=alias,
    )



import warnings
from typing import Any, Callable, Coroutine, Dict, List, Optional, Pattern, Tuple, Union

import asynctest
from httpcore import (
    AsyncByteStream,
    AsyncHTTPTransport,
    SyncByteStream,
    SyncHTTPTransport,
)

from .models import (
    URL,
    ContentDataTypes,
    Headers,
    HeaderTypes,
    Request,
    RequestPattern,
    Response,
    ResponseTemplate,
    TimeoutDict,
    build_request,
    build_response,
)


class BaseMockTransport:
    def __init__(
        self,
        assert_all_called: bool = True,
        assert_all_mocked: bool = True,
        base_url: Optional[str] = None,
    ) -> None:
        self._assert_all_called = assert_all_called
        self._assert_all_mocked = assert_all_mocked
        self._base_url = base_url

        self.patterns: List[RequestPattern] = []
        self.aliases: Dict[str, RequestPattern] = {}

        self.stats = asynctest.mock.MagicMock()
        self.calls: List[Tuple[Request, Optional[Response]]] = []

    def clear(self):
        """
        Clear added patterns.
        """
        self.patterns.clear()
        self.aliases.clear()

    def reset(self) -> None:
        """
        Resets call stats.
        """
        self.calls.clear()
        self.stats.reset_mock()
        for pattern in self.patterns:
            pattern.stats.reset_mock()

    def __getitem__(self, alias: str) -> Optional[RequestPattern]:
        return self.aliases.get(alias)

    def add(
        self,
        method: Union[str, Callable, RequestPattern],
        url: Optional[Union[str, Pattern]] = None,
        status_code: Optional[int] = None,
        content: Optional[ContentDataTypes] = None,
        content_type: Optional[str] = None,
        headers: Optional[HeaderTypes] = None,
        pass_through: bool = False,
        alias: Optional[str] = None,
    ) -> RequestPattern:
        """
        Adds a request pattern with given mocked response details.
        """
        if isinstance(method, RequestPattern):
            pattern = method

        else:
            response = ResponseTemplate(
                status_code, headers, content, content_type=content_type
            )
            pattern = RequestPattern(
                method,
                url,
                response,
                pass_through=pass_through,
                alias=alias,
                base_url=self._base_url,
            )

        self.patterns.append(pattern)
        if pattern.alias:
            self.aliases[pattern.alias] = pattern

        return pattern

    def get(
        self,
        url: Optional[Union[str, Pattern]] = None,
        status_code: Optional[int] = None,
        content: Optional[ContentDataTypes] = None,
        content_type: Optional[str] = None,
        headers: Optional[HeaderTypes] = None,
        pass_through: bool = False,
        alias: Optional[str] = None,
    ) -> RequestPattern:
        return self.add(
            "GET",
            url=url,
            status_code=status_code,
            content=content,
            content_type=content_type,
            headers=headers,
            pass_through=pass_through,
            alias=alias,
        )

    def post(
        self,
        url: Optional[Union[str, Pattern]] = None,
        status_code: Optional[int] = None,
        content: Optional[ContentDataTypes] = None,
        content_type: Optional[str] = None,
        headers: Optional[HeaderTypes] = None,
        pass_through: bool = False,
        alias: Optional[str] = None,
    ) -> RequestPattern:
        return self.add(
            "POST",
            url=url,
            status_code=status_code,
            content=content,
            content_type=content_type,
            headers=headers,
            pass_through=pass_through,
            alias=alias,
        )

    def put(
        self,
        url: Optional[Union[str, Pattern]] = None,
        status_code: Optional[int] = None,
        content: Optional[ContentDataTypes] = None,
        content_type: Optional[str] = None,
        headers: Optional[HeaderTypes] = None,
        pass_through: bool = False,
        alias: Optional[str] = None,
    ) -> RequestPattern:
        return self.add(
            "PUT",
            url=url,
            status_code=status_code,
            content=content,
            content_type=content_type,
            headers=headers,
            pass_through=pass_through,
            alias=alias,
        )

    def patch(
        self,
        url: Optional[Union[str, Pattern]] = None,
        status_code: Optional[int] = None,
        content: Optional[ContentDataTypes] = None,
        content_type: Optional[str] = None,
        headers: Optional[HeaderTypes] = None,
        pass_through: bool = False,
        alias: Optional[str] = None,
    ) -> RequestPattern:
        return self.add(
            "PATCH",
            url=url,
            status_code=status_code,
            content=content,
            content_type=content_type,
            headers=headers,
            pass_through=pass_through,
            alias=alias,
        )

    def delete(
        self,
        url: Optional[Union[str, Pattern]] = None,
        status_code: Optional[int] = None,
        content: Optional[ContentDataTypes] = None,
        content_type: Optional[str] = None,
        headers: Optional[HeaderTypes] = None,
        pass_through: bool = False,
        alias: Optional[str] = None,
    ) -> RequestPattern:
        return self.add(
            "DELETE",
            url=url,
            status_code=status_code,
            content=content,
            content_type=content_type,
            headers=headers,
            pass_through=pass_through,
            alias=alias,
        )

    def head(
        self,
        url: Optional[Union[str, Pattern]] = None,
        status_code: Optional[int] = None,
        content: Optional[ContentDataTypes] = None,
        content_type: Optional[str] = None,
        headers: Optional[HeaderTypes] = None,
        pass_through: bool = False,
        alias: Optional[str] = None,
    ) -> RequestPattern:
        return self.add(
            "HEAD",
            url=url,
            status_code=status_code,
            content=content,
            content_type=content_type,
            headers=headers,
            pass_through=pass_through,
            alias=alias,
        )

    def options(
        self,
        url: Optional[Union[str, Pattern]] = None,
        status_code: Optional[int] = None,
        content: Optional[ContentDataTypes] = None,
        content_type: Optional[str] = None,
        headers: Optional[HeaderTypes] = None,
        pass_through: bool = False,
        alias: Optional[str] = None,
    ) -> RequestPattern:
        return self.add(
            "OPTIONS",
            url=url,
            status_code=status_code,
            content=content,
            content_type=content_type,
            headers=headers,
            pass_through=pass_through,
            alias=alias,
        )

    def request(self, *args: Any, **kwargs: Any) -> RequestPattern:
        warnings.warn(
            "respx.request() is due to be deprecated. Use respx.add() instead.",
            DeprecationWarning,
            stacklevel=2,
        )
        return self.add(*args, **kwargs)

    def record(
        self,
        request: Any,
        response: Optional[Any],
        pattern: Optional[RequestPattern] = None,
    ) -> None:
        request = build_request(request)
        response = build_response(response, request=request)

        if pattern:
            pattern.stats(request, response)

        self.stats(request, response)

        # Copy stats due to unwanted use of property refs in the high-level api
        self.calls[:] = (
            (request, response) for (request, response), _ in self.stats.call_args_list
        )

    def assert_all_called(self) -> None:
        assert all(
            (pattern.called for pattern in self.patterns)
        ), "RESPX: some mocked requests were not called!"

    def match(
        self,
        method: bytes,
        url: URL,
        headers: Headers = None,
        stream: Union[SyncByteStream, AsyncByteStream] = None,
    ) -> Tuple[Optional[RequestPattern], Request, Optional[ResponseTemplate]]:
        request: Request = (method, url, headers, stream)

        matched_pattern: Optional[RequestPattern] = None
        matched_pattern_index: Optional[int] = None
        response: Optional[ResponseTemplate] = None

        for i, pattern in enumerate(self.patterns):
            match = pattern.match(request)
            if not match:
                continue

            if matched_pattern_index is not None:
                # Multiple matches found, drop and use the first one
                self.patterns.pop(matched_pattern_index)
                break

            matched_pattern = pattern
            matched_pattern_index = i

            if isinstance(match, ResponseTemplate):
                # Mock response
                response = match
            elif match == request:
                # Pass-through request
                response = None
            else:
                raise ValueError(
                    (
                        "Matched request pattern must return either a "
                        'ResponseTemplate or a Request, got "{}"'
                    ).format(type(match))
                )

        if matched_pattern is None:
            # Assert we always get a pattern match, if check is enabled
            assert not self._assert_all_mocked, f"RESPX: {request[1]!r} not mocked!"

            # Auto mock a successfull empty response
            response = ResponseTemplate()

        return matched_pattern, request, response

    def _sync_request(
        self,
        method: bytes,
        url: URL,
        headers: Headers = None,
        stream: SyncByteStream = None,
        timeout: TimeoutDict = None,
        pass_through: Callable[..., Response] = None,
    ) -> Response:
        pattern, request, response_template = self.match(method, url, headers, stream)

        try:
            if response_template is None:
                # Pass-through request
                if pass_through is None:
                    raise ValueError("pass_through not supported with manual transport")
                response = pass_through(method, url, headers, stream, timeout)
            else:
                response = response_template.raw
            return response
        except Exception:
            response = None
            raise
        finally:
            self.record(request, response, pattern=pattern)

    async def _async_request(
        self,
        method: bytes,
        url: URL,
        headers: Headers = None,
        stream: AsyncByteStream = None,
        timeout: TimeoutDict = None,
        pass_through: Callable[..., Coroutine[Any, Any, Response]] = None,
    ) -> Response:
        pattern, request, response_template = self.match(method, url, headers, stream)

        try:
            if response_template is None:
                # Pass-through request
                if pass_through is None:
                    raise ValueError("pass_through not supported with manual transport")
                response = await pass_through(method, url, headers, stream, timeout)
            else:
                response = await response_template.araw
            return response
        except Exception:
            response = None
            raise
        finally:
            self.record(request, response, pattern=pattern)

    def close(self) -> None:
        if self._assert_all_called:
            self.assert_all_called()

    async def aclose(self) -> None:
        self.close()


class SyncMockTransport(BaseMockTransport, SyncHTTPTransport):
    def request(  # type: ignore
        self,
        method: bytes,
        url: URL,
        headers: Headers = None,
        stream: SyncByteStream = None,
        timeout: TimeoutDict = None,
    ) -> Response:
        return self._sync_request(method, url, headers, stream, timeout)


class AsyncMockTransport(BaseMockTransport, AsyncHTTPTransport):
    async def request(  # type: ignore
        self,
        method: bytes,
        url: URL,
        headers: Headers = None,
        stream: AsyncByteStream = None,
        timeout: TimeoutDict = None,
    ) -> Response:
        return await self._async_request(method, url, headers, stream, timeout)
